

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use work.vlabifhw_pack.all;



entity test is
port (
	clk		        : in std_logic; 
    ser_txd			: out std_logic;
	ser_rxd			: in std_logic);
end test;

architecture rtl of test is

    -- Constants (constants.py)
    constant CPU_LEFT         : Natural := 31 ;
    constant SP_LEFT          : Natural := 10 ;
    constant DATA_LEFT        : Natural := 31 ;
    constant REF_LEFT         : Natural := 3; 
    constant ACTIVE_LEFT      : Natural := 23; 
    constant LOG2_NUM_CACHE_WORDS_PER_LINE : Natural := 2;
    constant LOG2_NUM_CACHE_LINES_PER_PAGE : Natural := 3;


    constant NUM_ENTRIES    : Natural := 2 ** ( REF_LEFT + 1 ) ;

    -- Internal links for debugging
    signal cpu_address         : std_logic_vector ( CPU_LEFT downto 0 );
    signal cpu_data_in         : std_logic_vector ( DATA_LEFT downto 0 );
                -- cpu_data_in is also used for "size" and "reference in"
    signal cpu_data_out        : std_logic_vector ( DATA_LEFT downto 0 );
                -- cpu_data_out is also used for "reference out"
    signal cpu_copy_loc        : std_logic_vector ( SP_LEFT downto 0 );
    signal cpu_size            : std_logic_vector ( SP_LEFT downto 0 );
    signal cpu_ben             : std_logic_vector ( 3 downto 0 );
    signal cpu_write           : std_logic; -- in 
    signal cpu_read            : std_logic; -- in 
    signal cpu_open            : std_logic; -- in 
    signal cpu_close           : std_logic; -- in 
    signal cpu_read_only       : std_logic; -- in 
    signal cpu_ack             : std_logic; -- out 
            -- Add Data LMB interface here
            -- Control lines from smmu to IPIF
    signal mst_cntl_rd_req     : std_logic ;
    signal mst_cntl_wr_req     : std_logic ;
    signal mst_cntl_bus_lock   : std_logic ;
    signal mst_cntl_burst      : std_logic ;
    signal mst_ip2bus_addr     : std_logic_vector ( CPU_LEFT downto 0 ) ;
    signal mst_ip2bus_be       : std_logic_vector ( 15 downto 0 ) ;
    signal mst_xfer_length     : std_logic_vector ( 11 downto 0 ) ;
    signal mst_go              : std_logic ;
            -- Status lines from IPIF to smmu
    signal mst_cmd_sm_set_done : std_logic ;
    signal mst_cmd_sm_set_error : std_logic ;
    signal mst_cmd_sm_set_timeout : std_logic ;
    signal mst_cmd_sm_busy     : std_logic ;
    signal mst_cmd_sm_clr_go   : std_logic ;
            -- Data interface
    signal output_next         : std_logic ;
    signal input_next          : std_logic ;
    signal data_output         : std_logic_vector ( DATA_LEFT downto 0 ) ;
    signal data_input          : std_logic_vector ( DATA_LEFT downto 0 ) ;
            -- Only service addresses matching this:
    signal match_mask          : std_logic_vector ( CPU_LEFT 
                                    downto ACTIVE_LEFT + 1 ) ;
    signal cache_active        : std_logic;
    signal cpu_cache_set_perm  : std_logic;
    signal cpu_cache_invalidate: std_logic;

    signal dbg_cpu_side_address     : std_logic_vector ( SP_LEFT downto 0 ) ;
    signal dbg_bus_side_address     : std_logic_vector ( SP_LEFT downto 0 ) ;
    signal dbg_cpu_side_data_in     : std_logic_vector ( DATA_LEFT downto 0 ) ;
    signal dbg_cpu_side_data_out    : std_logic_vector ( DATA_LEFT downto 0 ) ;
    signal dbg_bus_side_data_in     : std_logic_vector ( DATA_LEFT downto 0 ) ;
    signal dbg_bus_side_data_out    : std_logic_vector ( DATA_LEFT downto 0 ) ;
    signal dbg_cpu_side_write       : std_logic ;
    signal dbg_bus_side_write       : std_logic ;
    signal dbg_bus_side_cen         : std_logic ;
    signal dbg_match                : std_logic;
    signal dbg_match_in_read_only_area: std_logic;
    signal dbg_miss                 : std_logic;
    signal dbg_hit                  : std_logic;
    signal dbg_read_only_update_pos  : std_logic_vector ( REF_LEFT downto 0 );
    signal dbg_update_burst_boundary : std_logic;
    signal dbg_match_address        : std_logic_vector ( CPU_LEFT downto 0 ) ;
    signal dbg_code                 : std_logic_vector ( 7 downto 0 ) ;
    signal dbg_sp_cpu_valid         : std_logic;
    signal dbg_slow_ack             : std_logic;
    signal dbg_row_activation       : std_logic_vector ( 
                                            NUM_ENTRIES - 1 downto 0 ) ;
    signal dbg_match_ref            : std_logic_vector ( REF_LEFT downto 0 );
    signal dbg_tag_address          : std_logic_vector ( SP_LEFT - 1 downto 0 );
    signal dbg_permutation_in       : std_logic_vector ( SP_LEFT - 1 downto 0 );
    signal dbg_cache_hit            : std_logic;
    signal dbg_tag_in_16            : std_logic_vector ( 15 downto 0 );
    signal dbg_tag_out_16           : std_logic_vector ( 15 downto 0 );
    signal dbg_permutation_out      : std_logic_vector ( 15 downto 0 );
    signal dbg_cache_read           : std_logic;
    signal dbg_cache_write          : std_logic;
    signal dbg_enable_permutation   : std_logic;

    signal dc_out                   : std_logic;
    signal dc_in                    : std_logic;
    signal dc_control               : DC_Control_Wires;
    signal test_clock, test_reset   : std_logic;


begin

    -- The debugging hardware device
    vlhw : entity vlabifhw
        generic map (
            ext_channels => 1 ,
            fifo_depth => 1,
            baud_rate => 460800,
            clock_freq => 100000000 )
        port map (
            -- External hardware connections
            clk => clk,
            reset => '0',
            hw_tx => ser_txd,
            hw_rx => ser_rxd,

            -- Internal connections: these are not used in this example.
            out_channel_data => open,
            out_channel_wr => open,
            out_channel_rdy => "1",

            in_channel_data => x"00",
            in_channel_wr => "0",
            in_channel_rdy => open,

            -- Activation signal: not used
            active => open,

            -- Controls for device under test
            debug_clock => test_clock,
            debug_reset => test_reset,
            breakpoint => '0',

            dc_control => dc_control,
            dc_out => dc_out,
            dc_in => dc_in
        );

    -- Automatically generated component
    dc : entity Autogen_Debug_Entity
        port map (
            cache_active => cache_active,
            cpu_address => cpu_address,
            cpu_data_in => cpu_data_in,
            cpu_data_out => cpu_data_out,
            cpu_copy_loc => cpu_copy_loc,
            cpu_size => cpu_size,
            cpu_ben => cpu_ben,
            cpu_write => cpu_write,
            cpu_read => cpu_read,
            cpu_open => cpu_open,
            cpu_close => cpu_close,
            cpu_read_only => cpu_read_only,
            cpu_ack => cpu_ack,
            cpu_cache_set_perm => cpu_cache_set_perm,
            cpu_cache_invalidate => cpu_cache_invalidate,
            mst_cntl_rd_req => mst_cntl_rd_req,
            mst_cntl_wr_req => mst_cntl_wr_req,
            mst_cntl_bus_lock => mst_cntl_bus_lock,
            mst_cntl_burst => mst_cntl_burst,
            mst_ip2bus_addr => mst_ip2bus_addr,
            mst_ip2bus_be => mst_ip2bus_be,
            mst_xfer_length => mst_xfer_length,
            mst_go => mst_go,
            mst_cmd_sm_set_done => mst_cmd_sm_set_done,
            mst_cmd_sm_set_error => mst_cmd_sm_set_error,
            mst_cmd_sm_set_timeout => mst_cmd_sm_set_timeout,
            mst_cmd_sm_busy => mst_cmd_sm_busy,
            mst_cmd_sm_clr_go => mst_cmd_sm_clr_go,
            output_next => output_next,
            input_next => input_next,
            data_output => data_output,
            data_input => data_input,
            match_mask => match_mask,
            dbg_match_address => dbg_match_address,
            dbg_cpu_side_address => dbg_cpu_side_address,
            dbg_bus_side_address => dbg_bus_side_address,
            dbg_cpu_side_data_in => dbg_cpu_side_data_in,
            dbg_cpu_side_data_out => dbg_cpu_side_data_out,
            dbg_bus_side_data_in => dbg_bus_side_data_in,
            dbg_bus_side_data_out => dbg_bus_side_data_out,
            dbg_cpu_side_write => dbg_cpu_side_write,
            dbg_bus_side_write => dbg_bus_side_write,
            dbg_bus_side_cen => dbg_bus_side_cen,
            dbg_match => dbg_match,
            dbg_match_in_read_only_area => dbg_match_in_read_only_area,
            dbg_miss => dbg_miss,
            dbg_hit => dbg_hit,
            dbg_read_only_update_pos => dbg_read_only_update_pos,
            dbg_update_burst_boundary => dbg_update_burst_boundary,
            dbg_code => dbg_code,
            dbg_sp_cpu_valid => dbg_sp_cpu_valid,
            dbg_slow_ack => dbg_slow_ack,
            dbg_row_activation => dbg_row_activation,
            dbg_match_ref => dbg_match_ref,
            dbg_tag_address => dbg_tag_address,
            dbg_permutation_in => dbg_permutation_in,
            dbg_cache_hit => dbg_cache_hit,
            dbg_tag_in_16 => dbg_tag_in_16,
            dbg_tag_out_16 => dbg_tag_out_16,
            dbg_permutation_out => dbg_permutation_out,
            dbg_cache_read => dbg_cache_read,
            dbg_cache_write => dbg_cache_write,
            dbg_enable_permutation => dbg_enable_permutation,

            dc_control => dc_control,
            dc_in => dc_out,
            dc_out => dc_in ) ;

    -- Device under test
    smmu : entity smmu_core
        generic map (
            NUM_ENTRIES => NUM_ENTRIES,
            INITIATE_POINT => 1,
            CPU_LEFT => CPU_LEFT,
            SP_LEFT => SP_LEFT,
            DATA_LEFT => DATA_LEFT,
            REF_LEFT => REF_LEFT,
            LOG2_NUM_CACHE_WORDS_PER_LINE => LOG2_NUM_CACHE_WORDS_PER_LINE,
            LOG2_NUM_CACHE_LINES_PER_PAGE => LOG2_NUM_CACHE_LINES_PER_PAGE,
            NUM_READ_ONLY_ENTRIES => 3,
            ACTIVE_LEFT => ACTIVE_LEFT )
        port map (
            clk => test_clock,
            reset => test_reset,

            dbg_cpu_side_address => dbg_cpu_side_address,
            dbg_bus_side_address => dbg_bus_side_address,
            dbg_cpu_side_data_in => dbg_cpu_side_data_in,
            dbg_cpu_side_data_out => dbg_cpu_side_data_out,
            dbg_bus_side_data_in => dbg_bus_side_data_in,
            dbg_bus_side_data_out => dbg_bus_side_data_out,
            dbg_cpu_side_write => dbg_cpu_side_write,
            dbg_bus_side_write => dbg_bus_side_write,
            dbg_bus_side_cen => dbg_bus_side_cen,
            dbg_match => dbg_match,
            dbg_match_in_read_only_area => dbg_match_in_read_only_area,
            dbg_miss => dbg_miss,
            dbg_hit => dbg_hit,
            dbg_read_only_update_pos => dbg_read_only_update_pos,
            dbg_update_burst_boundary => dbg_update_burst_boundary,
            dbg_match_address => dbg_match_address,
            dbg_code => dbg_code,
            dbg_sp_cpu_valid => dbg_sp_cpu_valid,
            dbg_slow_ack => dbg_slow_ack,
            dbg_row_activation => dbg_row_activation,
            dbg_match_ref => dbg_match_ref,
            dbg_tag_address => dbg_tag_address,
            dbg_permutation_in => dbg_permutation_in,
            dbg_cache_hit => dbg_cache_hit,
            dbg_tag_in_16 => dbg_tag_in_16,
            dbg_tag_out_16 => dbg_tag_out_16,
            dbg_permutation_out => dbg_permutation_out,
            dbg_cache_read => dbg_cache_read,
            dbg_cache_write => dbg_cache_write,
            dbg_enable_permutation => dbg_enable_permutation,

            cache_active => cache_active,
            cpu_address => cpu_address,
            cpu_data_in => cpu_data_in,
            cpu_data_out => cpu_data_out,
            cpu_copy_loc => cpu_copy_loc,
            cpu_size => cpu_size,
            cpu_ben => cpu_ben,
            cpu_write => cpu_write,
            cpu_read => cpu_read,
            cpu_open => cpu_open,
            cpu_close => cpu_close,
            cpu_read_only => cpu_read_only,
            cpu_ack => cpu_ack,
            cpu_cache_set_perm => cpu_cache_set_perm,
            cpu_cache_invalidate => cpu_cache_invalidate,
            mst_cntl_rd_req => mst_cntl_rd_req,
            mst_cntl_wr_req => mst_cntl_wr_req,
            mst_cntl_bus_lock => mst_cntl_bus_lock,
            mst_cntl_burst => mst_cntl_burst,
            mst_ip2bus_addr => mst_ip2bus_addr,
            mst_ip2bus_be => mst_ip2bus_be,
            mst_xfer_length => mst_xfer_length,
            mst_go => mst_go,
            mst_cmd_sm_set_done => mst_cmd_sm_set_done,
            mst_cmd_sm_set_error => mst_cmd_sm_set_error,
            mst_cmd_sm_set_timeout => mst_cmd_sm_set_timeout,
            mst_cmd_sm_busy => mst_cmd_sm_busy,
            mst_cmd_sm_clr_go => mst_cmd_sm_clr_go,
            output_next => output_next,
            input_next => input_next,
            data_output => data_output,
            data_input => data_input,
            match_mask => match_mask);

end rtl;


