/*
 * Scratchpad MMU - SMMU simulator
 * Copyright (C) Jack Whitham 2009
 * http://www.jwhitham.org.uk/c/smmu.html
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation
 * (version 2.1 of the License only).
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <limits.h>


#include "smmu.h"
#include "mb_mem.h"

#define SMMU_VERSION    0x103

#define INVALID         UINT_MAX
#define TIMER_SANITY_LIMIT  0xf000


typedef struct SMMU_Table_Entry_struct
{
    unsigned                base_pointer ;
    unsigned                w_copy_location ;
    unsigned                w_size ;
    unsigned                min_address , max_address ;
    unsigned                acount ;
    unsigned                ref ;
    unsigned                opened_at ;
    char                    comment [ COMMENT_SIZE ] ;
    int                     in_use ;
} SMMU_Table_Entry ;

typedef struct SMMU_Context_struct {
    unsigned                check ;
    SMMU_Table_Entry *      table ;
    unsigned *              scratchpad_memory ;
    unsigned                num_ro_entries ;
    unsigned                num_rw_entries ;
    unsigned                num_entries ;
    unsigned                w_sp_size ;
    unsigned                num_open ;
    void *                  mem_context ;
    struct MB_System_Context_struct * sys_context ;
    MB_Context *            mb_context ;
    MB_Load_Fn              ifetch_fn ;
    MB_Load_Fn              load_fn ;
    MB_Store_Fn             store_fn ;
    unsigned                acount , miss_count , dbg_code ;
    unsigned                ctrl_mode ;
    unsigned                w_copy_loc , w_size ;
    unsigned                comment_ptr ;
    int                     offset_hint , w_read_only ;
    unsigned                result_counter , result_timer ;
    unsigned                log2_cache_lines_per_page;
    unsigned                log2_cache_words_per_line;
    unsigned                log2_cache_pages;
    unsigned *              cache_tags;
    unsigned *              cache_perm_table;
    unsigned                cache_enable, cache_ctrl_mode;
} SMMU_Context ;

typedef enum { SC_READ, SC_WRITE, 
            SC_INVALIDATE, SC_SET_PERMUTATION } Cache_Action;


static void SMMU_Store ( void * m_user , unsigned address , 
                    unsigned data , unsigned size ) ;
static void __STORE ( SMMU_Context * sc , unsigned address, 
                    unsigned data , unsigned size , unsigned limit ) ;
static unsigned __LOAD ( void * m_user , unsigned address , 
                    unsigned size , unsigned internal_mode ) ;
static unsigned SMMU_Load ( void * m_user , unsigned address , unsigned size ) ;
static unsigned SMMU_IFetch ( void * m_user , unsigned address , unsigned size ) ;
static void SMMU_Trace ( MB_Context * mc ,
                            MB_Trace_Name trace_name , const void * param ) ;
static unsigned Get_Config_Register (SMMU_Context * sc);
static unsigned Get_Config_Register_2 (SMMU_Context * sc);
static unsigned Config_Log2 (unsigned value);

SMMU_Context * SMMU_Init ( struct MB_System_Context_struct * sys ,
                unsigned num_rw_entries ,
                unsigned num_ro_entries ,
                unsigned w_sp_size )
{
    unsigned i ;
    SMMU_Context * sc = calloc ( 1 , sizeof ( SMMU_Context ) ) ;
    MB_Context * mc = MB_System_Get_MB_Context ( sys ) ;

    assert ( sc != NULL ) ;
    /* Initialise sc */
    sc -> num_entries = num_rw_entries + num_ro_entries + 1 ;
    sc -> num_rw_entries = num_rw_entries ;
    sc -> num_ro_entries = num_ro_entries ;
    sc -> table = calloc ( sizeof ( SMMU_Table_Entry ) , sc -> num_entries ) ;
    sc -> scratchpad_memory = calloc ( sizeof ( unsigned ) , w_sp_size ) ;
    assert ( sc -> table != NULL ) ;
    assert ( sc -> scratchpad_memory != NULL ) ;
    sc -> w_sp_size = w_sp_size ;
    sc -> mem_context = mc -> m_user ;
    sc -> mb_context = mc ;
    sc -> sys_context = sys ;
    for ( i = 0 ; i < sc -> num_entries ; i ++ )
    {
        sc -> table [ i ] . ref = i ;
    }

    /* Modify MB_Context */
    mc -> m_user = sc ;
    sc -> ifetch_fn = mc -> ifetch_fn ;
    mc -> ifetch_fn = SMMU_IFetch ;
    sc -> load_fn = mc -> load_fn ;
    mc -> load_fn = SMMU_Load ;
    sc -> store_fn = mc -> store_fn ;
    mc -> store_fn = SMMU_Store ;

    return sc ;
}

void SMMU_Cache_Init (SMMU_Context * sc,
                unsigned log2_cache_lines_per_page,
                unsigned log2_cache_words_per_line)
{
    unsigned    sp_bits, page_bits, tag_address_bits;

    assert(sc->cache_tags == NULL);
    assert(sc->cache_perm_table == NULL);
    sc->log2_cache_lines_per_page = log2_cache_lines_per_page;
    sc->log2_cache_words_per_line = log2_cache_words_per_line;
    sp_bits = Config_Log2(sc->w_sp_size) + 1;

    page_bits = log2_cache_lines_per_page + log2_cache_words_per_line;
    tag_address_bits = sp_bits - log2_cache_words_per_line;
    assert(sp_bits > page_bits);
    sc->log2_cache_pages = sp_bits - page_bits;
    sc->cache_tags = calloc(sizeof(unsigned), 1 << tag_address_bits);
    sc->cache_perm_table = calloc(sizeof(unsigned), 1 << sc->log2_cache_pages);
}

void SMMU_Delete ( SMMU_Context * sc )
{
    sc->mb_context->m_user = sc->mem_context ;
    free(sc->table);
    free(sc->scratchpad_memory);
    if (sc->cache_tags != NULL) free(sc->cache_tags);
    if (sc->cache_perm_table != NULL) free(sc->cache_perm_table);
    free ( sc ) ;
}

int SMMU_Inspect_Table ( SMMU_Context * sc ,
                unsigned ref ,
                unsigned * min_address,
                unsigned * max_address ,
                unsigned * base ,
                unsigned * opened_at ,
                char * comment_copy )
{
    if (( ref >= sc -> num_entries ) 
    || ( ! sc -> table [ ref ] . in_use ))
    {
        return 0 ;
    }
    if ( min_address != NULL )
        (* min_address) = sc -> table [ ref ] . min_address ;
    if ( max_address != NULL )
        (* max_address) = sc -> table [ ref ] . max_address ;
    if ( base != NULL )
        (* base) = sc -> table [ ref ] . base_pointer ;
    if ( opened_at != NULL )
        (* opened_at) = sc -> table [ ref ] . opened_at ;
    if ( comment_copy != NULL )
        strcpy ( comment_copy , sc -> table [ ref ] . comment ) ;

    return 1 ;
}

void SMMU_Dump_Table ( SMMU_Context * sc )
{
    unsigned i;

    for ( i = 0 ; i < sc -> num_entries ; i ++ )
    {
        if ( ! sc -> table[i].in_use )
        {
            printf("%3u: not in use\n" , i ) ;
        } else {
            printf("%3u: %08x:%08x -> %04x   %u\n" , i ,
                    sc -> table[i].min_address ,
                    sc -> table[i].max_address ,
                    sc -> table[i].w_copy_location ,
                    sc -> table[i].acount ) ;
        }
    }
}

unsigned SMMU_Open (
        SMMU_Context * sc ,
        unsigned notional_base_pointer ,
        unsigned w_size ,
        unsigned w_copy_location ,
        int offset_hint ,
        unsigned comment_ptr ,
        int read_only )
{
    SMMU_Table_Entry * found = NULL;
    unsigned ref = INVALID;
    unsigned i;
    unsigned cycle, cycle0 ;
    char buf [ NS ] ;

    sc -> dbg_code = 0xff;
    if (read_only)
    {
        for ( i = 1 ; i <= sc -> num_ro_entries ; i ++ )
        {
            if (! sc -> table[i].in_use)
            {
                found = & sc -> table[i];
                ref = i;
            }
        }
    } else {
        for ( i = sc -> num_ro_entries + 1 ; 
                i < ( sc -> num_rw_entries + sc -> num_ro_entries + 1 ) ; 
                i ++ )
        {
            if (! sc -> table[i].in_use)
            {
                found = & sc -> table[i];
                ref = i;
            }
        }
    }

    if (found == NULL) return INVALID;

    assert ( ref < sc -> num_entries ) ;
    assert ( ref != 0 ) ;

    assert (( w_copy_location + w_size ) <= sc -> w_sp_size );
    assert ( w_size != 0 ) ;

    cycle0 = cycle = sc -> mb_context -> clock_cycle ;

    for ( i = 0 ; i < w_size ; i ++ )
    {
        sc -> scratchpad_memory [ i + w_copy_location ] =
            htonl ( __LOAD ( sc , ( i << 2 ) + 
                            notional_base_pointer , 4 , 1 ) ) ;
        if ( i == 0 )
        {
            cycle = sc -> mb_context -> clock_cycle ;
        }
    }

    sc -> mb_context -> clock_cycle = cycle + ( i - 1 ) ;
    sc -> num_open ++ ;
    sc -> result_timer = sc -> mb_context -> clock_cycle - cycle0 ;
    assert (sc->result_timer <= TIMER_SANITY_LIMIT);
    sc -> result_counter ++ ;

    /* Initialise important stuff */
    found -> in_use = 1 ;
    found -> min_address = notional_base_pointer ;
    found -> max_address = notional_base_pointer + ( w_size << 2 ) ;
    found -> base_pointer = notional_base_pointer - offset_hint ;
    found -> w_size = w_size ;
    found -> w_copy_location = w_copy_location ;
    found -> acount = 0 ;

    /* Copy hints and comments */
    found -> comment [ 0 ] = '\0' ;
    if ( comment_ptr != 0 )
    {
        MB_Read ( sc -> sys_context , comment_ptr ,
                found -> comment , COMMENT_SIZE - 1 ) ;
        found -> comment [ COMMENT_SIZE - 1 ] = '\0' ;
    }
    found -> opened_at = sc -> mb_context -> cur_pc ;


    /* Trace output */
    snprintf ( buf , NS , 
            "%u: %08x:%08x -> %04x (%s)\n" , ref ,
                    found -> min_address ,
                    found -> max_address ,
                    found -> w_copy_location ,
                    found -> comment ) ;
    buf [ NS - 1 ] = '\0' ;
    SMMU_Trace ( sc -> mb_context , MB_OPEN , buf ) ;
    sc -> dbg_code = 0x9;
    return ref;
}

void SMMU_Close (
        SMMU_Context * sc ,
        unsigned ref )
{
    unsigned i, j;
    unsigned cycle, cycle0 ;
    unsigned copy_time = 1 ;
    SMMU_Table_Entry * found ;
    char buf [ NS ] ;
    
    sc -> dbg_code = 0xff;
    assert ( ref < sc -> num_entries ) ;
    assert ( ref != 0 ) ;

    found = & sc -> table [ ref ];
    assert ( found -> in_use ) ;

    cycle0 = cycle = sc -> mb_context -> clock_cycle ;
    if ( ref <= sc -> num_ro_entries )
    {
        /* Read-only entry. Discard. */    
        cycle = sc -> mb_context -> clock_cycle ;
    } else {
        /* Read-write entry. Copy to external memory. */
        for ( i = 0 ; i < found -> w_size ; i ++ )
        {
            unsigned data = ntohl ( sc -> scratchpad_memory [ i + 
                            found->w_copy_location ] ) ;
            unsigned address = ( i << 2 ) + found -> min_address ;

            copy_time ++ ;
            /* Copy to RW segment, or highest priority RO */
            __STORE ( sc , address , data , 4 , ref ) ;

            /* And one copy for each additional RO segment */
            for ( j = 1 ; j < sc -> num_ro_entries ; j ++ )
            {
                copy_time ++ ;
                __STORE ( sc , address , data , 4 , j + 1 ) ;
            }
            /* Store in external memory */
            if ( i == 0 )
            {
                cycle = sc -> mb_context -> clock_cycle ;
            }
            sc -> store_fn ( sc -> mem_context , address , data , 4 ) ;
        }
    }
    sc -> num_open -- ;
    sc -> mb_context -> clock_cycle = cycle + copy_time ;
    sc -> result_timer = sc -> mb_context -> clock_cycle - cycle0 ;
    assert (sc->result_timer <= TIMER_SANITY_LIMIT);
    sc -> result_counter ++ ;
    found -> in_use = 0;

    snprintf ( buf , NS , 
            "%u: %08x:%08x -> %04x (%s)\n" , ref ,
                    found -> min_address ,
                    found -> max_address ,
                    found -> w_copy_location ,
                    found -> comment ) ;
    buf [ NS - 1 ] = '\0' ;
    SMMU_Trace ( sc -> mb_context , MB_CLOSE , buf ) ;
    sc -> dbg_code = 0xc9;
}

static SMMU_Table_Entry * SMMU_Lookup ( SMMU_Context * sc , 
                unsigned address , unsigned limit )
{
    unsigned i;
    SMMU_Table_Entry * rc = NULL ;

    if ( sc -> num_open == 0 ) return NULL ;
    if (( address & ACTIVE_MASK ) != ACTIVE_AREA ) return NULL;

    for ( i = 0 ; i < limit ; i ++ )
    {
        if (( sc -> table[ i ] . in_use )
        && ( sc -> table[ i ].min_address <= address )
        && ( address < sc -> table[ i ].max_address ))
        {
            assert ( i != 0 ) ;
            rc = & sc -> table[i];
        }
    }
    return rc ;
}

static unsigned SW_Cache_Op(
        SMMU_Context * sc, unsigned address,
        unsigned data, unsigned size, Cache_Action action)
{
    unsigned line_within_page, page_nr, high_bits, tag, entry;
    unsigned ptr, permuted_page_nr, sp_address, cc;
    
    /* Get line nr */
    ptr = address >> (sc->log2_cache_words_per_line + 2); 

    /* Deconstruct pointer */
    line_within_page = ptr & ((1 << sc->log2_cache_lines_per_page) - 1);
    page_nr = (ptr >> sc->log2_cache_lines_per_page) & (
                                    (1 << sc->log2_cache_pages)- 1);
    high_bits = ptr >> (sc->log2_cache_lines_per_page +
                                    sc->log2_cache_pages);

    /* Apply permutation */
    permuted_page_nr = sc->cache_perm_table[ page_nr ];

    /* Identify tag and cache entry */
    tag = (high_bits << sc->log2_cache_pages) | page_nr;
    entry = (permuted_page_nr << sc->log2_cache_lines_per_page)
            | line_within_page;
    sp_address = (entry << (sc->log2_cache_words_per_line + 2)) |
            (address & ((1 << (sc->log2_cache_words_per_line + 2)) - 1));

    cc = sc->mb_context->clock_cycle;
    switch (action) {
        case SC_INVALIDATE:
            sc->cache_tags[ entry ] = 0;
            break;
        case SC_SET_PERMUTATION:
            sc->cache_perm_table[ page_nr ] = data;
            assert (data < (1 << sc->log2_cache_pages));
            break;
        case SC_WRITE:
            /* Write through */
            sc->store_fn(sc->mem_context, address, data, size);
            if (sc->cache_tags[ entry ] == tag) {
                /* hit: write to cache too */
                Page_Store(sc->scratchpad_memory,
                            sp_address, data, size);
            }
            data = 0;
            break;
        case SC_READ:
            /* Read */
            if (sc->cache_tags[ entry ] != tag) {
                /* miss: fill cache */
                unsigned i, size = 1 << sc->log2_cache_words_per_line;
                unsigned mask = ~((size << 2) - 1);
                unsigned sp_ptr = (sp_address & mask) >> 2;
                unsigned mem_ptr = address & mask;

                for (i = 0; i < size; i++) {
                    sc -> scratchpad_memory[ sp_ptr ] =
                        htonl(__LOAD(sc, mem_ptr, 4, 1));
                    if (i == (size - 2)) {
                        cc = sc -> mb_context -> clock_cycle - size;
                    }
                    mem_ptr += 4;
                    sp_ptr++;
                }
                sc->cache_tags[ entry ] = tag;
            }
            /* hit: read from cache */
            data = Page_Load(sc->scratchpad_memory, sp_address, size);
            break;
    }

    sc -> result_timer = sc -> mb_context -> clock_cycle - cc ;
    return data;
}

static unsigned __LOAD ( void * m_user , unsigned address , 
                    unsigned size , unsigned internal_mode )
{
    SMMU_Context * sc = (SMMU_Context *) m_user ;
    SMMU_Table_Entry * found ;

    sc -> result_counter ++ ;
    found = SMMU_Lookup ( sc , address , sc -> num_entries ) ;

    if ( found == NULL )
    {
        unsigned cc = sc -> mb_context -> clock_cycle ;
        unsigned out ;

        if (( ! internal_mode )
        && (( address & ACTIVE_MASK ) != SCRATCHPAD_AREA ))
        {
            sc -> miss_count ++ ;
            SMMU_Trace ( sc -> mb_context ,
                    MB_MISS_LOAD , IP ( address )) ;
        } 
        if ( sc->cache_enable
        && ( ! internal_mode )
        && (( address & ACTIVE_MASK ) == ACTIVE_AREA ))
        {
            return SW_Cache_Op(sc, address, 0, size, SC_READ);
        }

        out = sc -> load_fn ( sc -> mem_context , address , size ) ;
        sc -> result_timer = sc -> mb_context -> clock_cycle - cc ;
        assert (sc->result_timer <= TIMER_SANITY_LIMIT);
        return out ;
    } else {
        if ( ! internal_mode )
        {
            found -> acount ++ ;
            sc -> acount ++ ;
            SMMU_Trace ( sc -> mb_context ,
                    MB_HIT_LOAD , IP ( found -> ref )) ;
        }
        address += ( found -> w_copy_location << 2 )
                - found -> min_address ;
        assert (( address >> 2 ) < sc -> w_sp_size ) ;
        sc -> result_timer = 1 ;
        return Page_Load ( sc -> scratchpad_memory ,
                            address , size ) ;
    }
} 

static unsigned SMMU_Load ( void * m_user , unsigned address , unsigned size )
{
    SMMU_Context * sc = (SMMU_Context *) m_user ;

    if (( address & ACTIVE_MASK ) == CONTROL_AREA )
    {
        /* Access status registers */
        assert ( size == 4 ) ;
        switch (( address & 0xff ) >> 2 )
        {
            case 0 :    assert (sc->result_timer <= TIMER_SANITY_LIMIT);
                        return (( sc -> result_timer & 0xffff ) | 
                            ( sc -> result_counter << 16 )) ;
            case 1 :    return sc -> mb_context -> clock_cycle ;
            case 2 :    return sc -> acount ;
            case 3 :    return sc -> miss_count ;
            case 4 :    return 0xdbdb0000U | (sc -> dbg_code & 0xffff);
            case 6 :    return Get_Config_Register_2 ( sc );
            case 7 :    return Get_Config_Register ( sc );
            default :   assert ( 0 ) ;
                        break ;
        }
        return 0 ;
    }
    if ( sc -> ctrl_mode )
    {
        if ( sc -> cache_ctrl_mode )
        {
            /* Load in cache control mode = invalidate line */
            SW_Cache_Op(sc, address, 0, 0, SC_INVALIDATE);
            sc -> ctrl_mode = 0 ;
            sc -> cache_ctrl_mode = 0 ;
            return 0;
        } else {
            /* Load in control mode = Open */
            unsigned out ;
            sc -> ctrl_mode = 0 ;
            out = SMMU_Open ( sc , address , sc -> w_size , sc -> w_copy_loc ,
                sc -> offset_hint , sc -> comment_ptr , sc -> w_read_only ) ;
            sc -> offset_hint = 0 ;
            sc -> comment_ptr = 0 ;
            return out ;
        }
    }
    return __LOAD ( m_user , address , size , 0 ) ; 
}

static void __STORE ( SMMU_Context * sc , unsigned in_address, 
        unsigned data , unsigned size , unsigned limit )
{
    SMMU_Table_Entry * found = SMMU_Lookup ( sc , in_address , limit ) ;
    unsigned internal_mode = ( limit != sc -> num_entries ) ;
    unsigned address ;

    if ( ! internal_mode )
    {
        sc -> result_counter ++ ;
    }

    if (( found == NULL )                           /* Miss */
    || ( found -> ref <= sc -> num_ro_entries ))    /* Store on Read-only */
    {
        unsigned cc = sc -> mb_context -> clock_cycle ;
        if (( ! internal_mode )
        && (( in_address & ACTIVE_MASK ) != SCRATCHPAD_AREA ))
        {
            sc -> miss_count ++ ;
            SMMU_Trace ( sc -> mb_context ,
                    MB_MISS_STORE , IP ( in_address )) ;
        }
        if ( sc->cache_enable
        && ( ! internal_mode )
        && (( in_address & ACTIVE_MASK ) == ACTIVE_AREA ))
        {
            SW_Cache_Op(sc, in_address, data, size, SC_WRITE);
            return;
        }
        if ( ! internal_mode )
        {
            sc -> store_fn ( sc -> mem_context , in_address , data , size ) ;
            sc -> result_timer = sc -> mb_context -> clock_cycle - cc;
            assert (sc->result_timer <= TIMER_SANITY_LIMIT);
        }
    }
    
    if ( found != NULL )                            /* Hit */
    {
        if ( ! internal_mode )
        {
            found -> acount ++ ;
            sc -> acount ++ ;
            SMMU_Trace ( sc -> mb_context ,
                    MB_HIT_STORE , IP ( found -> ref )) ;
        }
        assert ( found -> ref != 0 ) ;
        address = in_address + ( found -> w_copy_location << 2 )
                - found -> min_address ;
        assert (( address >> 2 ) < sc -> w_sp_size ) ;
        Page_Store ( sc -> scratchpad_memory ,
                            address , data , size ) ;

        if ( found -> ref <= sc -> num_ro_entries )
        {
            /* Store on read-only. Check lower-priority read-only
             * entries are also up to date. */
            unsigned i ;
            for ( i = 1 ; i < found -> ref ; i ++ )
            {
                SMMU_Table_Entry * f2 = SMMU_Lookup ( sc , 
                                in_address , i + 1 ) ;
                if ( f2 != NULL )
                {
                    address = in_address + 
                            ( f2 -> w_copy_location << 2 )
                            - f2 -> min_address ;
                    assert (( address >> 2 ) < sc -> w_sp_size ) ;
                    Page_Store ( sc -> scratchpad_memory ,
                                        address , data , size ) ;
                }
            }
        } else {
            sc -> result_timer = 1 ;
        }
    }
} 

static void SMMU_Store ( void * m_user , unsigned address , 
                    unsigned data , unsigned size )
{
    SMMU_Context * sc = (SMMU_Context *) m_user ;

    if (( address & ACTIVE_MASK ) == CONTROL_AREA )
    {
        assert ( size == 4 ) ;
        switch (( address & 0xff ) >> 2 )
        {
            case 0 :
                /* Enter control mode - program copy_loc and size registers */
                sc -> ctrl_mode = 1 ;
                sc -> cache_ctrl_mode = 0;
                sc -> w_copy_loc = data & 0xffff ;
                sc -> w_size = ( data >> 16 ) & 0x7fff ;
                sc -> w_read_only = data >> 31 ;
                assert (( sc -> w_size + sc -> w_copy_loc ) 
                                <= sc -> w_sp_size ) ;
                break ;
            case 1 :
                /* write offset hint - not required */
                sc -> offset_hint = (int) data ;
                break ;
            case 2 :
                /* write comment pointer hint - not required */
                sc -> comment_ptr = data ;
                break ;
            case 3 :
                /* cache control */
                if (data & 2) {
                    /* Enter cache control mode */
                    sc -> ctrl_mode = 1;
                    sc -> cache_ctrl_mode = 1;
                }
                /* Enable/disable cache */
                sc -> cache_enable = (data & 1) &&
                        (sc -> log2_cache_lines_per_page != 0);
                break;
            default :   
                assert ( 0 ) ;
                break ;
        }
        return ;
    }
    if ( sc -> ctrl_mode )
    {
        if ( sc -> cache_ctrl_mode )
        {
            /* Store in cache control mode = set permutation */
            SW_Cache_Op(sc, address, data, 4, SC_SET_PERMUTATION);
            sc -> ctrl_mode = 0 ;
            sc -> cache_ctrl_mode = 0 ;
            return;
        } else {
            /* Store in control mode = Close */
            sc -> ctrl_mode = 0 ;
            SMMU_Close ( sc , data ) ;
            return ;
        }
    }
    __STORE ( sc , address , data , size , sc -> num_entries ) ;
}

static unsigned SMMU_IFetch ( void * m_user , unsigned address , unsigned size )
{
    SMMU_Context * sc = (SMMU_Context *) m_user ;
    return sc -> ifetch_fn ( sc -> mem_context , address , size ) ;
}

static void SMMU_Trace ( MB_Context * mc ,
                            MB_Trace_Name trace_name , const void * param )
{
    return mc -> trace_fn ( mc -> t_user , mc , trace_name , param ) ;
}

static unsigned Config_Log2 (unsigned value)
{
    unsigned i;
    for (i = 1; i < 16; i++)
    {
        unsigned test = 1 << i;
        if (value <= test) {
            return i - 1;
        }
    }
    return 15;
}

static unsigned Get_Config_Register (SMMU_Context * sc)
{
    unsigned out;
    unsigned active_left = ACTIVE_BITS - 1;

    out = SMMU_VERSION << 20;
    out |= Config_Log2 ( sc -> num_entries ) << 16;
    out |= (active_left & 0xf) << 12;
    out |= (sc -> num_ro_entries & 0xf) << 8;
    out |= Config_Log2 ( sc -> w_sp_size ) << 4;
    out |= 0x8; /* Extra scratchpad available */
    if (sc -> log2_cache_lines_per_page != 0) {
        out |= 0x4; /* Cache feature available */
    }
    return out;
}

static unsigned Get_Config_Register_2 (SMMU_Context * sc)
{
    unsigned out = 0;

    out |= (sc->log2_cache_lines_per_page - 1) << 28;
    out |= (sc->log2_cache_words_per_line - 1) << 24;
    return out;
}



