/*
 * Scratchpad MMU - Test program module 2 - Cache tests
 * Copyright (C) Jack Whitham 2009
 * http://www.jwhitham.org.uk/c/smmu.html
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation
 * (version 2.1 of the License only).
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */
#undef NDEBUG
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <assert.h>
#include "print.h"

#ifdef __MICROBLAZE__
#include "mb_interface.h"
#include "xparameters.h"
#include "hwdriver.h"
#endif

#define NUM_ELEMENTS        0x4000
#define SMALL_NUMBER        1381
#define TEST_MARGIN         9

#define HIT_LATENCY         2
#define EXTRA_CHECKING

unsigned rng ( void ) ;
void reset_rng ( void ) ;
int braindamage_test ( void );

static int Ascending (const void * a, const void * b) 
{
    unsigned a1 = ((unsigned *) a)[ 0 ];
    unsigned b1 = ((unsigned *) b)[ 0 ];

    if (a1 < b1) {
        return -1;
    } else if (a1 > b1) {
        return 1;
    }
    return 0;
}

static unsigned elementA[ NUM_ELEMENTS ];
static unsigned elementB[ NUM_ELEMENTS ];

/* Enable cache, write, disable cache.
 * Byte, half word, or word */
unsigned writemod1(unsigned ctrl,
         unsigned value0, unsigned ptr, unsigned decision);
unsigned writemod2(unsigned ctrl,
         unsigned value0, unsigned ptr, unsigned decision);
unsigned writemod4(unsigned ctrl,
         unsigned value0, unsigned ptr, unsigned decision);


static int SW_Cache_Op(
        unsigned * tag_table,
        unsigned * perm_table,
        unsigned ptr,
        int invalidate)
{
    unsigned line_within_page, page_nr, high_bits, tag, entry;
    unsigned permuted_page_nr;
    
    /* Get line nr */
    ptr = ptr >> (LOG2_NUM_CACHE_WORDS_PER_LINE + 2); 

    /* Deconstruct pointer */
    line_within_page = ptr & (NUM_CACHE_LINES_PER_PAGE - 1);
    page_nr = (ptr >> LOG2_NUM_CACHE_LINES_PER_PAGE) & (
                                    NUM_CACHE_PAGES - 1);
    high_bits = ptr >> (LOG2_NUM_CACHE_LINES_PER_PAGE +
                                    LOG2_NUM_CACHE_PAGES);

    /* Apply permutation */
    permuted_page_nr = perm_table[ page_nr ];

    /* Identify tag and cache entry */
    tag = (high_bits << LOG2_NUM_CACHE_PAGES) | page_nr;
    entry = (permuted_page_nr << LOG2_NUM_CACHE_LINES_PER_PAGE)
            | line_within_page;

    if (invalidate) {
        tag_table[ entry ] = 0;
        return 0; /* no action */
    } else if (tag_table[ entry ] == tag) {
        return 1; /* hit */
    } else {
        tag_table[ entry ] = tag;
        return 0; /* miss */
    }
}


static int Cache_Blaster(unsigned key, unsigned limit, 
                unsigned test_code, unsigned * perm_table)
{
    unsigned        i, j, size, ec = 0;
    unsigned        ptr;
    unsigned *      ctrl = (unsigned *) CONTROL;
    volatile unsigned *elementB4 = 
                        (unsigned *) ((void *) elementB);
    volatile unsigned short *elementB2 = 
                        (unsigned short *) ((void *) elementB);
    volatile unsigned char * elementB1 = 
                        (unsigned char *) ((void *) elementB);
    unsigned *      tag_table;
    unsigned        log2_num_lines = LOG2_NUM_CACHE_PAGES + 
                        LOG2_NUM_CACHE_LINES_PER_PAGE;
    unsigned        num_lines = 1 << log2_num_lines;
    unsigned        hit_count = 0;
    unsigned        miss_count = 0;

    tag_table = alloca(4 * num_lines);
    memset(tag_table, 0, 4 * num_lines);

    for (j = 0; j < NUM_ELEMENTS; j++) {
        elementA[ j ] = j | (0xff << 24); //rng();
        elementB4[ j ] = elementA[ j ] ^ key ;
    }
    for (i = 0; (i < limit) && (ec < 30); i ++) {
        unsigned    value0, value1, value2;
        unsigned    debug, status, decision;
        int         hit;

        /* Read test */
        j = (rng() + i) % NUM_ELEMENTS;
        //j = i % NUM_ELEMENTS;
        ptr = (unsigned) &elementA[ j ];
        hit = SW_Cache_Op(tag_table, perm_table, ptr, 0);

        value0 = value1 = 0;
        debug = status = 0;
        asm volatile (
                "lwi        %1, %3, 0x0\n"      /* Read without cache */
                "ori        r18, r0, 1\n"
                "swi        r18, %2, 0xc\n"     /* Cache enable */
                "lwi        %0, %3, 0x0\n"     /* Read with cache */
                "lwi        %5, %2, 0x0\n"      /* Status code */
                "lwi        %4, %2, 0x10\n"     /* Debug code */
                "swi        r0, %2, 0xc\n"      /* Cache disable */
            : "=&r"(value0), "=&r"(value1), "=&r"(ctrl), "=&r"(ptr),
                "=&r"(debug), "=&r"(status)
            : "0"(value0), "1"(value1), "2"(ctrl), "3"(ptr),
                "4"(debug), "5"(status) );

        value2 = elementB4[ j ];
        if ((value0 != value1)
        || ((value0 ^ key ^ value2) != 0))
        {
            print("\rRead reject i=");
            printhex(i);
            print(" ptr=");
            printhex(ptr);
            print(" v0=");
            printhex(value0);
            print(" v1=");
            printhex(value1);
            print(" v2^K=");
            printhex(value2 ^ key);
#ifdef EXTRA_CHECKING
            print("\r\ndebug=");
            printhex(debug);
            print(" status=");
            printhex(status);
#endif
            print("\r\n");
            ec ++;
        }
        /* Was it a cache hit? Should it have been? */
        status &= 0xffff;
        if (hit) {
            /* Should be hit */
            if (status > HIT_LATENCY) {
                print("\rRead should have been hit (but wasn't).\r\n");
                printhex(status);
                print(" ");
                printhex(debug);
                print("\r\n");
                ec ++;
            } 
            hit_count ++;
        } else {
            /* Should be miss */
            if (status <= HIT_LATENCY) {
                print("\rRead should have been miss (but wasn't).\r\n");
                printhex(status);
                print(" ");
                printhex(debug);
                print("\r\n");
                ec ++;
            } 
            miss_count ++;
        }

        /* Write test */
        j = rng();
        //j = 99;
        decision = j & 0x7;
        size = (j >> 3) & 0x3;
        if (decision <= 3) {
#ifdef EXTRA_CHECKING
            unsigned k, left, right, fail;
#endif
            unsigned old = 0, debug = 0;

            /* Decision:
                0       Write through (don't ensure data is cached first)
                1       Write through (one load operation first)
                2       Write (no cache)
                3       Write through (two load operations first)
            */
            j = j >> 5;
            if ((size == 3) || (size == 0)) {
                size = 4;
            }
            value0 = ~i; //rng();
            switch (size) {
                case 4 :    j %= NUM_ELEMENTS;
                            ptr = ((unsigned) elementA) + (j * 4);
                            old = elementB4[ j ] ^ key;
                            elementB4[ j ] = value0 ^ key;
                            debug = writemod4((unsigned)CONTROL, value0,
                                    ptr, decision);
                            break;
                case 2 :    j %= NUM_ELEMENTS * 2;
                            value0 &= 0xffff;
                            ptr = ((unsigned) elementA) + (j * 2);
                            elementB4[ j / 2 ] ^= key;
                            old = elementB4[ j / 2 ];
                            elementB2[ j ] = value0;
                            elementB4[ j / 2 ] ^= key;
                            j /= 2;
                            debug = writemod2((unsigned)CONTROL, value0,
                                    ptr, decision);
                            break;
                case 1 :    j %= NUM_ELEMENTS * 4;
                            value0 &= 0xff;
                            ptr = ((unsigned) elementA) + j;
                            elementB4[ j / 4 ] ^= key;
                            old = elementB4[ j / 4 ];
                            elementB1[ j ] = value0;
                            elementB4[ j / 4 ] ^= key;
                            j /= 4;
                            debug = writemod1((unsigned)CONTROL, value0,
                                    ptr, decision);
                            break;
                default :   assert(0);
                            break;
            }
            if ((decision == 1) || (decision == 3)) {
                /* Cache access */
                SW_Cache_Op(tag_table, perm_table, ptr, 0);
            } else if (decision == 2) {
                /* Invalidate cache line */
                SW_Cache_Op(tag_table, perm_table, ptr, 1);
            }
    
            value1 = value0;


#ifdef EXTRA_CHECKING
            if (j > TEST_MARGIN) {
                left = j - TEST_MARGIN;
            } else {
                left = 0;
            }
            if (j < (NUM_ELEMENTS - TEST_MARGIN)) {
                right = j + TEST_MARGIN;
            } else {
                right = NUM_ELEMENTS;
            }
            fail = 0;
            for (k = left; k < right; k++) {
                if ((elementA[ k ] ^ key) != elementB4[ k ]) {
                    fail = 1;
                }
            }
            if (fail) {
                for (k = left; k < right; k++) {
                    printhex(k);
                    print(" = ");
                    printhex(elementA[ k ]);
                    print(" ");
                    printhex(elementB4[ k ] ^ key);
                    if (k == j) {
                        print(" <==");
                    }
                    print("\r\n");
                }
                print("Write reject i=");
                printhex(i);
                print(" value0=");
                printhex(value0);
                print(" value1=");
                printhex(value1);
                print("\r\ndbg_code=");
                printhex(debug);
                print(" old=");
                printhex(old);
                print("\r\n");
                ec ++;
            }
#endif
        } 

        /* Status update */
        if ((i & 0x0fff) == 0x0fff) {
            print("\r");
            printhex(i | test_code);
        }
    }
    if (ec != 0) {
        print("Errors detected. Halt\r\n");
        assert(0);
    }
    print(" hits ");
    printhex(hit_count);
    print(" misses ");
    printhex(miss_count);

    /* Final check */
    i = braindamage_test();
    if (i != 0) {
        print("Brain damage detected! (1)\r\n");
        assert(0);
    }
    print("\r");
    HW_Cache_Enable(1); /* Also flushes the cache */
    print("..");
    i = braindamage_test();
    if (i != 0) {
        print("Brain damage detected! (2)\r\n");
        assert(0);
    }
    print("  ");
    HW_Cache_Disable();
    print("..");
    HW_Cache_Enable(1); 
    print("  ");
    HW_Cache_Disable();
    return 1;
}

void Cache_Test ()
{
    unsigned i,j;
    unsigned test_code = 0;
    const unsigned TCA = 0x1 << 20;
    const unsigned RUN_TO = 0x1 << 17;
    unsigned * perm_table;
    unsigned * do_not_touch;
    unsigned page_size = SP_SIZE / NUM_CACHE_PAGES;

    if (!HAS_CACHE) return;

    printf ( "Cache test:\r\n" ) ;

    do_not_touch = alloca(4 * (SP_SIZE - page_size));
    perm_table = alloca(4 * NUM_CACHE_PAGES);
    for (i = 0; i < NUM_CACHE_PAGES; i++) {
        perm_table[ i ] = i;
    }

    /* First stage - simple stuff, random change and test */
    for (i = 0; i < 31; i ++) {
        unsigned dnt_ref = 0;
        unsigned check = 0;
        /* Generate new permutation */
        switch (i) {
            case 0:     
                break; /* No new permutation */
            case 2:
            case 9:
                /* Everything mapped to 1 cache line */
                for (j = 0; j < NUM_CACHE_PAGES; j++) {
                    perm_table[ j ] = 0;
                    HW_Cache_Set_Permutation (j, perm_table[ j ]);
                }
                /* Remaining space must not be touched */
                dnt_ref = HW_OPEN(do_not_touch, 
                        4 * (SP_SIZE - page_size), page_size);
                assert(dnt_ref < NUM_ENTRIES);
                assert(dnt_ref > 0);
                check = rng();
                for (j = 0; j < (SP_SIZE - page_size); j ++) {
                    do_not_touch[ j ] = check + j;
                }
                break;
            default:
                /* Random permutation */
                for (j = 0; j < NUM_CACHE_PAGES; j++) {
                    perm_table[ j ] = rng() % NUM_CACHE_PAGES;
                    HW_Cache_Set_Permutation (j, perm_table[ j ]);
                }
                break;
        }

        Cache_Blaster(rng(), RUN_TO, test_code, perm_table);
        test_code += TCA;

        if (dnt_ref != 0) {
            /* Check untouched space */
            for (j = 0; j < (SP_SIZE - page_size); j ++) {
                if (do_not_touch[ j ] != (check + j)) {
                    print("\r\nCorruption detected at scratchpad address ");
                    printhex(page_size + j);
                    print("\r\n");
                    assert(0);
                }
            }
            HW_CLOSE(dnt_ref);
        }
    }
    for (i = 0; i < NUM_CACHE_PAGES; i++) {
        /* Return to default permutation */
        HW_Cache_Set_Permutation (i, i);
    }

    /* Second stage - full code execution and sorting */
    for (i = 0; i < 0x200; i ++) {
        if ((i & 0xf) == 0xf) {
            print("\r");
            printhex(i | test_code);
        }
        switch (i & 31) {
            case 0 :
            case 10 :   HW_Cache_Disable();
                        break;
            case 2 :
            case 11 :   for (j = 0; j < SMALL_NUMBER; j++) {
                            elementA[ j ] = rng();
                        }
                        break;
            case 1 :
            case 12 :   HW_Cache_Enable(1);
                        break;
            default :   break ;
        }

        elementA[ 0 ] = i;
        elementA[ 1 ] = i + 1;
        qsort(elementA, SMALL_NUMBER, sizeof(unsigned), Ascending);
        assert (elementA[ SMALL_NUMBER - 1 ] != 0);
        assert (elementA[ SMALL_NUMBER - 1 ] > elementA[ 0 ]);
        for (j = 1; j < SMALL_NUMBER; j++) {
            assert (elementA[ j ] >= elementA[ j - 1 ]);
        }
        for (j = 0; j < SMALL_NUMBER; j += 4) {
            elementA[ j + 1 ] = rng();
            elementA[ j + 3 ] = ((SMALL_NUMBER - j) * 0x1337) % 1021;
        }
        elementA[ 9 ] = i;
        elementA[ 8 ] = i + 1;
        qsort(elementA, SMALL_NUMBER, sizeof(unsigned), Ascending);
        assert (elementA[ SMALL_NUMBER - 1 ] != 0);
        assert (elementA[ SMALL_NUMBER - 1 ] > elementA[ 0 ]);
        for (j = 1; j < SMALL_NUMBER; j++) {
            assert (elementA[ j ] >= elementA[ j - 1 ]);
        }
        j = 1 + (rng() % ((SMALL_NUMBER / 2) - 2));
        memmove(&elementA[ 0 ], &elementA[ SMALL_NUMBER / 2 ],
                j * sizeof(unsigned));
        elementA[ 4 ] = i;
        elementA[ 9 ] = i + 1;
        qsort(elementA, SMALL_NUMBER, sizeof(unsigned), Ascending);
        assert (elementA[ SMALL_NUMBER - 1 ] != 0);
        assert (elementA[ SMALL_NUMBER - 1 ] > elementA[ 0 ]);
        for (j = 1; j < SMALL_NUMBER; j++) {
            assert (elementA[ j ] >= elementA[ j - 1 ]);
        }
    }
    HW_Cache_Disable();
    print("\rCache tested ok\r\n");
}

